
from distutils.spawn import find_executable
from sys import version_info
from subprocess import Popen, PIPE
from hashlib import sha256
from os import remove, SEEK_END, SEEK_SET, statvfs, urandom


class NoOpenSSLError(ImportError):
	pass


class EncryptionError(Exception):
	pass


def file_hash(pth):
	"""
	Calculate the checksum of a file; return length-40 binary that includes the algorithm.
	"""
	shaer = sha256()
	with open(pth, 'rb') as fh:
		data = ' '
		while data:
			data = fh.read(32768)
			shaer.update(data)
	return b'sha256__' + shaer.digest()


def check_prereq():
	if version_info < (3, 0):
		raise SystemError('fileenc needs python3')
	if not find_executable('openssl'):
		raise NoOpenSSLError('fileenc needs openssl to be installed, but it was not found\n' +
			'on Ubuntu: sudo apt-get install openssl\notherwise check: https://www.openssl.org/community/binaries.html')


def shred_file(pth, *, rounds=3):
	"""
	Try to shred a file with shred command, otherwise just overwrite it (may not work due to things like buffering).
	"""
	success = False
	if find_executable('shred'):
		proc = Popen([
			'shred', '-n', '{0:d}'.format(rounds), '-z', pth,
		], stdout=PIPE, stderr=PIPE)
		out, err = proc.communicate()
		if not proc.returncode and not err:
			success = True
	if not success:
		overwrite_file(pth)
	remove(pth)


def overwrite_file(pth, *, rounds=3):
	"""
	Custom version of shred, as a fallback.
	"""
	blocksize = statvfs(pth).f_bsize
	with open(pth, 'rb+') as fh:
		assert fh.seekable(), 'cannot overwrite non-seekable file'
		fh.seek(0, SEEK_END)
		upto = fh.tell()
		for r in range(rounds):
			fh.seek(0, SEEK_SET)
			for k in range(0, upto, blocksize):
				fh.write(urandom(blocksize))
			fh.flush()






from argparse import ArgumentParser
from filecmp import cmp
from getpass import getpass
from glob import glob
from os import getpid
from tempfile import gettempdir
from shutil import move
from sys import stderr, argv
from os.path import join, isfile
from .enc_dec import stretch_key, encrypt_file, decrypt_file
from .misc import check_prereq, EncryptionError, overwrite_file, shred_file

check_prereq()


def handle_cmds(args):
	parser = ArgumentParser(prog='fileenc', description='Encrypt and decrypt files by pattern.',
		epilog='Note: `filedec` is the same as `fileenc -d` ; Info at https://github.com/mverleg/fileenc_openssl')

	parser.add_argument('-k', '--key', dest='key', action='store', type=str, default=None,
		help='the key to use for encryption; you will be prompted for one if this is not provided (more secure)')
	parser.add_argument('-i', '--input', dest='inp', action='store', type=str, required=False,
		help='input file, directory or pattern as a single string (required for encrypting; defaults to *.enc when decrypting)')
	parser.add_argument('-o', '--output', dest='outp', action='store', type=str, default='',
		help='optionally, output file or directory; .enc will be appended to each file')
	parser.add_argument('-d', '--decrypt', dest='encrypt', action='store_false',
		help='decrypt the input file(s) (as opposed to encrypt, which is the default)')
	parser.add_argument('-f', '--overwrite', dest='overwrite', action='store_true',
		help='overwrite existing files when decrypting (encrypting always overwrites)')
	parser.add_argument('-r', '--remove', dest='remove', action='store_true',
		help='shred the input file after en/decrypting (after --check)')
	parser.add_argument('-c', '--check', dest='test', action='store_true',
		help='test the encryption by reversing it (abort on failure) (only for ENcryption due to salting)')
	parser.add_argument('-1', '--once', dest='key_once', action='store_true',
		help='prompt for the key only once (only applicable if --key and --decrypt are not set)')

	args = parser.parse_args(args)

	if args.test and not args.encrypt:
		print('--check ignored for decryption')

	if args.inp:
		files = glob(args.inp)
	elif not args.encrypt:
		files = glob('*.enc')
	else:
		stderr.write('--input not provided')
		exit(3)
	if not files:
		stderr.write('no files found that match "{0:s}"\n'.format(args.inp))
		exit(1)

	key = args.key
	if not key:
		try:
			key = getpass(prompt='key: ')
			if args.key_once and args.encrypt:
				key_repeat = getpass(prompt='repeat: ')
				assert key == key_repeat
		except KeyboardInterrupt:
			print('aborted')
			exit(1)
	try:
		stretched_key = stretch_key(key)

		for file in files:
			tmp_pth = join(args.outp, '{0:s}.tmp'.format(file))
			if args.encrypt:
				encrypt_file(file, encpth=tmp_pth, key=stretched_key)
				to_file = join(args.outp, '{0:s}.enc'.format(file))
				print(file, '->', to_file)
			else:
				decrypt_file(file, rawpth=tmp_pth, key=stretched_key)
				to_file = join(args.outp, (file[:-4] if file.endswith('.enc') else file))
				print(to_file, '<-', file)
			if isfile(to_file):
				if args.overwrite or args.encrypt:
					shred_file(to_file)
				else:
					raise IOError(('a file exists at the target destination "{0:s}" '
						'(use --overwrite/-f to overwrite it)').format(to_file))
			if args.test:
				check_pth = join(gettempdir(), 'endfile-check-{0:d}.tmp'.format(getpid()))
				if args.encrypt:
					decrypt_file(tmp_pth, rawpth=check_pth, key=stretched_key)
					if cmp(file, check_pth):
						print(' tested', file)
					else:
						raise EncryptionError('checking "{0:s}" with --check did not yield identical file ("{1:s}")'
							.format(file, check_pth))
					shred_file(check_pth)
			move(tmp_pth, to_file)
			if args.remove:
				print(' removing', file)
				shred_file(file)
	except (EncryptionError, IOError) as err:
		# stderr.write('encrypt/decrypt error!\n')
		stderr.write(str(err) + '\n')
		exit(2)


def handle_cmds_encrypt():
	return handle_cmds(argv[1:])


def handle_cmds_decrypt():
	return handle_cmds(['--decrypt'] + argv[1:])


if __name__ == '__main__':
	handle_cmds_encrypt()






from subprocess import Popen, PIPE
from base64 import urlsafe_b64encode
from hashlib import sha256
from os import SEEK_END
from re import match
from .misc import check_prereq, EncryptionError, file_hash


check_prereq()


def stretch_key(key, *, rounds=86198):  # python3 required
	"""
	Apply a hash to a string for many rounds. The default takes about 0.1s on a simple 2014 CPU,
	which is probably enough of a deterrent if the key is not terrible.
	"""
	assert rounds >= 1
	if not match(r'^[\w!@#$%^&*()+-=~\[\]{};:`\'"|/?,.<>\\]{4,}$', key):
		raise EncryptionError('invalid key; keys should consist of at least four ascii letters, '
			'numbers or characters among !@#$%^&*()_+-=~[]{};:`\'"|/?,.<>\\')
	binkey = key.encode('ascii')
	for rnd in range(rounds):
		shaer = sha256()
		shaer.update(binkey)
		binkey = shaer.digest()
	return urlsafe_b64encode(binkey).decode('ascii')[:-1]


def encrypt_file(rawpth, *, encpth=None, key):
	"""
	Use openssl to encrypt a file using `aes-256` with a salt (no key stretching implicit).
	"""
	if encpth is None:
		encpth = '{0:s}.enc'.format(rawpth)
	proc = Popen([
		'openssl', 'aes-256-cbc', '-salt',
		'-in', rawpth, '-out', encpth,
		'-e', '-k', '{0:s}'.format(key),
	], stdout=PIPE, stderr=PIPE)
	out, err = proc.communicate()
	if err:
		raise EncryptionError('encrypting "{0:s}" failed due to openssl error:\n"{1:s}"'
			.format(rawpth, err.decode('ascii').strip()))
	checksum = file_hash(rawpth)
	with open(encpth, 'ab') as fh:
		fh.write(b'Checksum_' + checksum)
	return encpth


def decrypt_file(encpth, *, rawpth=None, key):
	"""
	Reverse of `encrypt_file`.
	"""
	if rawpth is None:
		rawpth = encpth
		if rawpth.endswith('.enc'):
			rawpth = rawpth[:-4]
	with open(encpth, 'rb') as fh:
		fh.seek(-49, SEEK_END)
		if not fh.read(9) == b'Checksum_':
			raise EncryptionError('no checksum found at the end of "{0:s}"'.format(encpth))
		checksum_found = fh.read()
	with open(encpth, 'ab') as fh:
		fh.seek(-49, SEEK_END)
		fh.truncate()
	proc = Popen([
		'openssl', 'aes-256-cbc', '-salt',
		'-in', encpth, '-out', rawpth,
		'-d', '-k', '{0:s}'.format(key),
	], stdout=PIPE, stderr=PIPE)
	out, err = proc.communicate()
	with open(encpth, 'ab') as fh:
		fh.seek(0, SEEK_END)
		fh.write(b'Checksum_' + checksum_found)
	if err:
		raise EncryptionError('decrypting "{0:s}" failed due to openssl error:\n"{1:s}"'
			.format(encpth, err.decode('ascii').strip()))
	checksum_decrypted = file_hash(rawpth)
	if not checksum_found == checksum_decrypted:
		raise EncryptionError('The decrypted file does not have the same checksum as the original!\n' +
			' original:  {0:}\n decrypted: {1:}\n'.format(checksum_found, checksum_decrypted))
	return rawpth





# -*- coding: utf-8 -*-

from sys import version_info
from setuptools import setup


if version_info < (3, 0):
	raise SystemError('fileenc-openssl only works with python3')


with open('README.rst', 'r') as fh:
	readme = fh.read()

setup(
	name='fileenc-openssl',
	description='allows one to easily encrypt and decrypt files symmetrically using openssl and python3',
	long_description=readme,
	url='https://github.com/mverleg/fileenc_openssl',
	author='Mark V',
	maintainer='(the author)',
	author_email='mdilligaf@gmail.com',
	license='Revised BSD License (LICENSE.txt)',
	keywords=['encryption', 'openssl',],
	version='1.2.4',
	packages=['fileenc_openssl'],
	include_package_data=True,
	zip_safe=False,
	entry_points={
		'console_scripts': [
			'fileenc=fileenc_openssl.commands:handle_cmds_encrypt',
			'filedec=fileenc_openssl.commands:handle_cmds_decrypt',
		],
	},
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Natural Language :: English',
		'License :: OSI Approved :: BSD License',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
	],
	install_requires=[
	],
)


